@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -51,6 +51,9 @@
 #include "uinput.h"
 
 #include "../src/storage.h"
+#include "../src/manager.h"
+#include "../src/dbus-common.h"
+#include "adapter.h"
 
 #include "device.h"
 #include "error.h"
@@ -306,7 +309,7 @@ failed:
 static inline DBusMessage *not_supported(DBusMessage *msg)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-							"Not suported");
+							"Not supported");
 }
 
 static inline DBusMessage *in_progress(DBusMessage *msg)
@@ -334,7 +337,6 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	struct input_device *idev = iconn->idev;
 	struct fake_input *fake = iconn->fake;
 	DBusMessage *reply;
-	const char *path;
 
 	if (err < 0)
 		goto failed;
@@ -360,13 +362,6 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	reply = dbus_message_new_method_return(iconn->pending_connect);
 	g_dbus_send_message(idev->conn, reply);
 
-	/* Sending the Connected signal */
-	path = dbus_message_get_path(iconn->pending_connect);
-	g_dbus_emit_signal(idev->conn, path,
-			INPUT_DEVICE_INTERFACE, "Connected",
-			DBUS_TYPE_STRING, &iconn->uuid,
-			DBUS_TYPE_INVALID);
-
 	dbus_message_unref(iconn->pending_connect);
 	iconn->pending_connect = NULL;
 
@@ -399,14 +394,13 @@ static gboolean intr_watch_cb(GIOChannel *chan, GIOCondition cond, gpointer data
 {
 	struct input_conn *iconn = data;
 	struct input_device *idev = iconn->idev;
+	gboolean connected = FALSE;
 
 	if (cond & (G_IO_HUP | G_IO_ERR))
 		g_io_channel_close(chan);
 
-	g_dbus_emit_signal(idev->conn, idev->path,
-			INPUT_DEVICE_INTERFACE, "Disconnected",
-			DBUS_TYPE_STRING, &iconn->uuid,
-			DBUS_TYPE_INVALID);
+	emit_property_changed(idev->conn, idev->path, INPUT_DEVICE_INTERFACE,
+				"Connected", DBUS_TYPE_BOOLEAN, &connected);
 
 	g_source_remove(iconn->ctrl_watch);
 	iconn->ctrl_watch = 0;
@@ -425,14 +419,10 @@ static gboolean intr_watch_cb(GIOChannel *chan, GIOCondition cond, gpointer data
 static gboolean ctrl_watch_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 {
 	struct input_conn *iconn = data;
-	struct input_device *idev = iconn->idev;
 
 	if (cond & (G_IO_HUP | G_IO_ERR))
 		g_io_channel_close(chan);
 
-	g_dbus_emit_signal(idev->conn, idev->path, INPUT_DEVICE_INTERFACE,
-			"Disconnected", DBUS_TYPE_INVALID);
-
 	g_source_remove(iconn->intr_watch);
 	iconn->intr_watch = 0;
 	iconn->ctrl_watch = 0;
@@ -472,86 +462,6 @@ static int fake_hid_disconnect(struct input_conn *iconn)
 	return fhid->disconnect(iconn->fake);
 }
 
-static int encrypt_link(const char *src_addr, const char *dst_addr)
-{
-	char filename[PATH_MAX + 1];
-	struct hci_conn_info_req *cr;
-	int dd, err, dev_id;
-	auth_requested_cp cp;
-	char *str;
-
-	create_name(filename, PATH_MAX, STORAGEDIR, src_addr, "linkkeys");
-
-	str = textfile_get(filename, dst_addr);
-	if (!str) {
-		error("Encryption link key not found");
-		return -ENOKEY;
-	}
-
-	free(str);
-
-	cr = g_try_malloc0(sizeof(*cr) + sizeof(struct hci_conn_info));
-	if (!cr)
-		return -ENOMEM;
-
-	dev_id = hci_devid(src_addr);
-	if (dev_id < 0) {
-		g_free(cr);
-		return -errno;
-	}
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0) {
-		g_free(cr);
-		return -errno;
-	}
-
-	str2ba(dst_addr, &cr->bdaddr);
-	cr->type = ACL_LINK;
-
-	if (ioctl(dd, HCIGETCONNINFO, (unsigned long) cr) < 0)
-		goto fail;
-
-	if (cr->conn_info->link_mode & HCI_LM_ENCRYPT) {
-		/* Already encrypted */
-		goto done;
-	}
-
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(cr->conn_info->handle);
-
-	if (hci_send_cmd(dd, OGF_LINK_CTL, OCF_AUTH_REQUESTED,
-						sizeof(cp), &cp) < 0) {
-		error("Link authentication failed: %s (%d)",
-						strerror(errno), errno);
-		goto fail;
-	}
-
-#if 0
-	/* FIXME: This needs to be done after auth completed */
-	if (hci_encrypt_link(dd, handle, 1, 5000) < 0) {
-		error("Link encryption failed: %s (%d)",
-						strerror(errno), errno);
-		goto fail;
-	}
-#endif
-
-done:
-	g_free(cr);
-
-	hci_close_dev(dd);
-
-	return 0;
-
-fail:
-	g_free(cr);
-
-	err = errno;
-	hci_close_dev(dd);
-
-	return -err;
-}
-
 static void epox_endian_quirk(unsigned char *data, int size)
 {
 	/* USAGE_PAGE (Keyboard)	05 07
@@ -634,75 +544,121 @@ static void extract_hid_record(sdp_record_t *rec, struct hidp_connadd_req *req)
 	}
 }
 
-static int hidp_connadd(const bdaddr_t *src, const bdaddr_t *dst, int ctrl_sk,
-		int intr_sk, int timeout, const char *name, const uint32_t handle)
+static int ioctl_connadd(struct hidp_connadd_req *req)
 {
-	struct hidp_connadd_req req;
+	int ctl, err = 0;
+
+	ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HIDP);
+	if (ctl < 0)
+		return -errno;
+
+	if (ioctl(ctl, HIDPCONNADD, req) < 0)
+		err = errno;
+
+	close(ctl);
+
+	return -err;
+}
+
+static void encrypt_completed(uint8_t status, gpointer user_data)
+{
+	struct hidp_connadd_req *req = user_data;
+	int err;
+
+	if (status) {
+		error("Encryption failed: %s(0x%x)",
+				strerror(bt_error(status)), status);
+		goto failed;
+	}
+
+	err = ioctl_connadd(req);
+	if (err == 0)
+		goto cleanup;
+
+	error("ioctl_connadd(): %s(%d)", strerror(-err), -err);
+failed:
+	close(req->intr_sock);
+	close(req->ctrl_sock);
+
+cleanup:
+	if (req->rd_data)
+		free(req->rd_data);
+
+	g_free(req);
+}
+
+static int hidp_add_connection(const struct input_device *idev,
+			       const struct input_conn *iconn)
+			       
+{
+	struct hidp_connadd_req *req;
 	struct fake_hid *fake_hid;
 	struct fake_input *fake;
 	sdp_record_t *rec;
 	char src_addr[18], dst_addr[18];
-	int ctl, err;
+	int err;
 
-	memset(&req, 0, sizeof(req));
-	req.ctrl_sock = ctrl_sk;
-	req.intr_sock = intr_sk;
-	req.flags     = 0;
-	req.idle_to   = timeout;
+	req = g_new0(struct hidp_connadd_req, 1);
+	req->ctrl_sock = iconn->ctrl_sk;
+	req->intr_sock = iconn->intr_sk;
+	req->flags     = 0;
+	req->idle_to   = iconn->timeout;
 
-	ba2str(src, src_addr);
-	ba2str(dst, dst_addr);
+	ba2str(&idev->src, src_addr);
+	ba2str(&idev->dst, dst_addr);
 
-	rec = fetch_record(src_addr, dst_addr, handle);
+	rec = fetch_record(src_addr, dst_addr, idev->handle);
 	if (!rec) {
 		error("Rejected connection from unknown device %s", dst_addr);
 		err = -EPERM;
 		goto cleanup;
 	}
 
-	extract_hid_record(rec, &req);
+	extract_hid_record(rec, req);
 	sdp_record_free(rec);
 
-	fake_hid = get_fake_hid(req.vendor, req.product);
+	read_device_id(src_addr, dst_addr, NULL,
+				&req->vendor, &req->product, &req->version);
+
+	fake_hid = get_fake_hid(req->vendor, req->product);
 	if (fake_hid) {
 		fake = g_new0(struct fake_input, 1);
 		fake->connect = fake_hid_connect;
 		fake->disconnect = fake_hid_disconnect;
 		fake->priv = fake_hid;
-		return fake_hid_connadd(fake, intr_sk, fake_hid);
-	}
-
-	ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HIDP);
-	if (ctl < 0) {
-		error("Can't open HIDP interface");
-		return -errno;
-	}
-
-	if (req.subclass & 0x40) {
-		err = encrypt_link(src_addr, dst_addr);
-		if (err < 0 && err != -EACCES)
-			goto cleanup;
-	}
-
-	if (name)
-		strncpy(req.name, name, 128);
-
-	if (req.vendor == 0x054c && req.product == 0x0268) {
-		unsigned char buf[] = { 0x53, 0xf4,  0x42, 0x03, 0x00, 0x00 };
-		err = write(ctrl_sk, buf, sizeof(buf));
-	}
-
-	err = ioctl(ctl, HIDPCONNADD, &req);
-	if (err < 0) {
-		close(ctl);
+		err = fake_hid_connadd(fake, iconn->intr_sk, fake_hid);
 		goto cleanup;
 	}
 
-	info("New input device %s (%s)", dst_addr, req.name);
+	if (idev->name)
+		strncpy(req->name, idev->name, 128);
+
+	/* Encryption is mandatory for keyboards */
+	if (req->subclass & 0x40) {
+		err = bt_acl_encrypt(&idev->src, &idev->dst, encrypt_completed, req);
+		if (err == 0) {
+			/* Waiting async encryption */
+			return 0;
+		} else if (err != -EALREADY) {
+			error("bt_acl_encrypt(): %s(%d)", strerror(-err), -err);
+			goto cleanup;
+		}
+
+		/* Link already encrypted - reset error */
+		err = 0;
+	}
+
+	if (req->vendor == 0x054c && req->product == 0x0268) {
+		unsigned char buf[] = { 0x53, 0xf4,  0x42, 0x03, 0x00, 0x00 };
+		err = write(iconn->ctrl_sk, buf, sizeof(buf));
+	}
+
+	err = ioctl_connadd(req);
 
 cleanup:
-	if (req.rd_data)
-		free(req.rd_data);
+	if (req->rd_data)
+		free(req->rd_data);
+	g_free(req);
 
 	return err;
 }
@@ -720,18 +676,13 @@ static void interrupt_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	}
 
 	iconn->intr_sk = g_io_channel_unix_get_fd(chan);
-	err = hidp_connadd(&idev->src, &idev->dst,
-				iconn->ctrl_sk, iconn->intr_sk,
-				iconn->timeout, idev->name, idev->handle);
+	err = hidp_add_connection(idev, iconn);
+
 	if (err < 0)
 		goto failed;
 
 	iconn->intr_watch = create_watch(iconn->intr_sk, intr_watch_cb, iconn);
 	iconn->ctrl_watch = create_watch(iconn->ctrl_sk, ctrl_watch_cb, iconn);
-	g_dbus_emit_signal(idev->conn, idev->path,
-			INPUT_DEVICE_INTERFACE, "Connected",
-			DBUS_TYPE_STRING, &iconn->uuid,
-			DBUS_TYPE_INVALID);
 
 	/* Replying to the requestor */
 	g_dbus_send_reply(idev->conn, iconn->pending_connect, DBUS_TYPE_INVALID);
@@ -755,6 +706,7 @@ static void control_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 {
 	struct input_conn *iconn = user_data;
 	struct input_device *idev = iconn->idev;
+	DBusMessage *reply;
 
 	if (err < 0) {
 		error("connect(): %s (%d)", strerror(-err), -err);
@@ -776,7 +728,8 @@ static void control_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 failed:
 	iconn->ctrl_sk = -1;
-	connection_attempt_failed(iconn->pending_connect, -err);
+	reply = connection_attempt_failed(iconn->pending_connect, -err);
+	g_dbus_send_message(idev->conn, reply);
 	dbus_message_unref(iconn->pending_connect);
 	iconn->pending_connect = NULL;
 }
@@ -981,27 +934,6 @@ static DBusMessage *device_disconnect(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusMessage *device_is_connected(DBusConnection *conn,
-						DBusMessage *msg, void *data)
-{
-	struct input_device *idev = data;
-	dbus_bool_t connected = FALSE;
-	GSList *l;
-
-	for (l = idev->connections; l; l = l->next) {
-		struct input_conn *iconn = l->data;
-
-		if (!is_connected(iconn))
-			continue;
-
-		connected = TRUE;
-		break;
-	}
-
-	return g_dbus_create_reply(msg, DBUS_TYPE_BOOLEAN, &connected,
-							DBUS_TYPE_INVALID);
-}
-
 static void device_unregister(void *data)
 {
 	struct input_device *idev = data;
@@ -1014,17 +946,53 @@ static void device_unregister(void *data)
 	input_device_free(idev);
 }
 
+static gint connected_cmp(gpointer a, gpointer b)
+{
+	struct input_conn *iconn = a;
+
+	return !is_connected(iconn);
+}
+
+static DBusMessage *device_get_properties(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct input_device *idev = data;
+	DBusMessage *reply;
+	DBusMessageIter iter;
+	DBusMessageIter dict;
+	dbus_bool_t connected;
+
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return NULL;
+
+	dbus_message_iter_init_append(reply, &iter);
+
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+	/* Connected */
+	connected = !!g_slist_find_custom(idev->connections, NULL,
+					(GCompareFunc) connected_cmp);
+	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &connected);
+
+	dbus_message_iter_close_container(&iter, &dict);
+
+	return reply;
+}
+
 static GDBusMethodTable device_methods[] = {
 	{ "Connect",		"",	"",	device_connect,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Disconnect",		"",	"",	device_disconnect	},
-	{ "IsConnected",	"",	"b",	device_is_connected	},
+	{ "GetProperties",	"",	"a{sv}",device_get_properties },
 	{ }
 };
 
 static GDBusSignalTable device_signals[] = {
-	{ "Connected",		"ss"	},
-	{ "Disconnected",	"s"	},
+	{ "PropertyChanged",	"sv"	},
 	{ }
 };
 
@@ -1228,6 +1196,7 @@ int input_device_connadd(const bdaddr_t *src, const bdaddr_t *dst)
 	struct input_device *idev;
 	struct input_conn *iconn;
 	int err;
+	gboolean connected;
 
 	idev = find_device(src, dst);
 	if (!idev)
@@ -1237,17 +1206,16 @@ int input_device_connadd(const bdaddr_t *src, const bdaddr_t *dst)
 	if (!iconn)
 		return -ENOENT;
 
-	err = hidp_connadd(src, dst, iconn->ctrl_sk, iconn->intr_sk,
-				iconn->timeout, idev->name, idev->handle);
+	err = hidp_add_connection(idev, iconn);
 	if (err < 0)
 		goto error;
 
 	iconn->intr_watch = create_watch(iconn->intr_sk, intr_watch_cb, iconn);
 	iconn->ctrl_watch = create_watch(iconn->ctrl_sk, ctrl_watch_cb, iconn);
-	g_dbus_emit_signal(idev->conn, idev->path,
-			INPUT_DEVICE_INTERFACE, "Connected",
-			DBUS_TYPE_STRING, &iconn->uuid,
-			DBUS_TYPE_INVALID);
+	connected = TRUE;
+	emit_property_changed(idev->conn, idev->path, INPUT_DEVICE_INTERFACE,
+				"Connected", DBUS_TYPE_BOOLEAN, &connected);
+
 	return 0;
 
 error:
@@ -1255,5 +1223,6 @@ error:
 	close(iconn->intr_sk);
 	iconn->ctrl_sk = -1;
 	iconn->intr_sk = -1;
+
 	return err;
 }
